/* File: w2c.c
 * Conversion from 44.1kHz 16bit mono WAV file to CAS format.
 * Input is read from standard input, output is sent to standard output.
 * Written by Csaba Markus <csaba.markus.at.freemail.hu>
 * March 2005, Budapest.
 *
 * Note: This program has been written for and run on Linux. DOS/Windows
 * line termination handling may require to reopen stdout in binary mode
 * if someone ever tries to run it on those "OS"s.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXSAMPLES (4*1024*1024)
#define WINDOWSIZE 89
#define PULSEWIDTH 20
#define BIT1LIMIT 66
#define SILENCELIMIT 150
#define PULSE_SEPARATION 30

short buffer[MAXSAMPLES];       /* Array of 16-bit values */

int main(int argc, char **argv)
{
    int p, n;           /* index and number of samples in buffer */
    int ap = 0;         /* Absolute Position in input file */
    int lspap = 0;      /* Last Start Pulse Absolute Position */
    int lpap = 0;       /* Last Pulse Absolute Position */
    int ad;             /* current Average Difference */
    int ads[WINDOWSIZE];/* Average Difference values for most recent samples*/
    int locmax;         /* value of local maximum */
    int lplocmax = 0;   /* local maximum at last pulse */
    int sample;         /* value of current sample */
    int bit = 0;        /* most recent data bit */
    unsigned char byte = 0xff;  /* current data byte (accumulating) */
    int bits;           /* Number of bits accumulated in current byte */
    int synching = 1;   /* true when looking for A5H after zero bytes */
    int synchnotzero = 100;     /* number of bits read since last zero byte */
    int blockno = 0;    /* Serial number of the current block */
    int finished = 0;   /* indicates if all is done */

    if (argc > 1) {
        fprintf(stderr, "Usage: %s <inputfile.wav >outputfile.cas\n"
                "where inputfile.wav is a 44.1kHz 16bit mono WAV file.\n",
                argv[0]);
        return 1;
    }

    fread(buffer, 60, 1, stdin);
    if (feof(stdin)) {
        fprintf(stderr, "Error: too short input WAV file");
        return 1;
    }

    memset(buffer, 0, 2 * WINDOWSIZE);
    p = n = WINDOWSIZE;

    for (;;) {
        if (p >= n) {
            if (finished) {
                fprintf(stderr, "Conversion finished.\n");
                return 0;
            }
            p -= n - WINDOWSIZE;
            memmove(buffer, buffer + n - WINDOWSIZE, WINDOWSIZE * 2);
            n = fread(buffer + WINDOWSIZE, 2 /* sizeof(short) */,
                      MAXSAMPLES - WINDOWSIZE, stdin);
            if (n == 0) {
                memset(buffer + p, 0, 4 * SILENCELIMIT);
                n = 2 * SILENCELIMIT;
                finished = 1;
            }
            n += WINDOWSIZE;
        }
        sample = buffer[p];
        ad = abs(sample - buffer[p - 3]);
        ads[ap % WINDOWSIZE] = ad;

        if (ap - lpap > PULSE_SEPARATION) {
            int i;
            locmax = ads[(ap - PULSEWIDTH / 2) % WINDOWSIZE];
            for (i = 0; i < PULSEWIDTH; i++) {
                if (ads[(ap - i) % WINDOWSIZE] > locmax) {
                    locmax = 0;
                    break;
                }
            }
        }
        else {
            locmax = 0;
        }

        if (locmax > 500) {
            if (locmax >= lplocmax / 3) {       /* pulse found */
                lplocmax = locmax;
                lpap = ap;
                if (ap - lspap < BIT1LIMIT) {
                    bit = 1;
                }
                else {                          /* start pulse found */
                    lspap = ap;
                    byte = (byte << 1) | bit;
                    bit = 0;
                    if (synching) {
                        if (byte == 0xa5 && synchnotzero < 10) {
                            /* block start found */
                            int i;
                            for (i = 0; i < 255; i++) {
                                putchar('\0');
                            }
                            putchar(0xa5);
                            synching = 0;
                            bits = 0;
                            fprintf(stderr, "*** SYNCH at %d, block %d ***\n", ap,  ++blockno);
                        }
                        else {
                            if (byte == 0) {
                                synchnotzero = 0;
                            }
                            else {
                                synchnotzero++;
                            }
                        }
                    }
                    else {
                        if (++bits == 8) {
                            putchar(byte);
                            bits = 0;
                        }
                    }
                }
            }
        }
        p++;
        ap++;
        if (!synching && ap - lspap > SILENCELIMIT) {
            if (bits == 7) {
                byte = (byte << 1) | bit;
                putchar(byte);
            }
            synching = 1;
            synchnotzero = 100;
            lplocmax = 0;
            fprintf(stderr, "*** SILENCE at %d ***\n", lspap);
        }
    }

    return 0;
}
